/*
 * Copyright (C) 2004-2006  Autodesk, Inc.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser
 * General Public License as published by the Free Software Foundation.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

//#include "stdafx.h"
#include "Geometry.h"
#include <string.h>

/// Enumeration of possible byte orders.
enum ByteOrder
{
    eWkbXDR = 0, // Big Endian
    eWkbNDR = 1  // Little Endian
};

/// Enumeration of various PostGIS-specific flags stored in EWKB data.
enum GeometryTypeFlags
{
    eWkbFlagZ    = 0x80000000,
    eWkbFlagM    = 0x40000000,
    eWkbFlagSrid = 0x20000000,
    eWkbFlagBBox = 0x10000000
};

enum PostGISGeometryType
{
    PostGISGeometryType_Curve           = 8,
    PostGISGeometryType_CompoundCurve   = 9,
    PostGISGeometryType_CurvePolygon    = 10,
    PostGISGeometryType_MultiCurve      = 11,
    PostGISGeometryType_MultiSurface    = 12
};

static unsigned long GetOrdinatesFromDimension(unsigned long dim);

static void FgfFromExtendedWkb(FdoByteArray*& fgfBytes,
                                       unsigned char* ewkb,
                                       unsigned long size,
                                       unsigned long& currentByte);

static void ExtendedWkbFromFgf(FdoByteArray*& ewkbBytes,
                                       FdoByte* fgf,
                                       unsigned long size,
                                       unsigned long srid,
                                       unsigned long& currentByte,
                                       unsigned long& totalDims);

static void CopyGeomFromExtendedWkb(FdoByteArray*& dest,
                                       FdoByte* src,
                                       unsigned long size,
                                       unsigned long& currentByte,
                                       unsigned long geomType,
                                       unsigned long geomDim);

static void CopyGeomFromFgf(FdoByteArray*& dest,
                                       FdoByte* src,
                                       unsigned long size,
                                       unsigned long& currentByte,
                                       unsigned long geomType,
                                       unsigned long geomDim,
                                       unsigned long& tarGeomType);

static void PostGISCircularStringToFdoCurve(FdoByteArray*& dest,
                                       FdoByte* src,
                                       unsigned long size,
                                       unsigned long& currentByte,
                                       unsigned long geomDim);

static void PostGISCompoundCurveToFdoCurve(FdoByteArray*& dest,
                                       FdoByte* src,
                                       unsigned long size,
                                       unsigned long& currentByte,
                                       unsigned long geomDim);

static void FdoCurveToPostGISCompoundCurve(FdoByteArray*& dest,
                                       FdoByte* src,
                                       unsigned long size,
                                       unsigned long& currentByte,
                                       unsigned long geomDim);

void *GeometryFromWkb (void *wkb, unsigned long size)
{
    FdoPtr<FdoFgfGeometryFactory> factory;
    FdoByteArray* array;
    FdoIGeometry *g = NULL;

    factory = FdoFgfGeometryFactory::GetInstance ();
    array = FdoByteArray::Create ((FdoByte*)wkb, size);
    try
    {
        g = factory->CreateGeometryFromWkb (array);
        array->Release ();
    }
    catch (...)
    {
        array->Release ();
    }

    return (void *)g;
}

void FreeGeometry (void*& geometry)
{
	FdoIGeometry *g = (FdoIGeometry *)geometry;

    FDO_SAFE_RELEASE(g);
}

char *WkbFromGeometry (void *geometry, int *size)
{
    FdoPtr<FdoFgfGeometryFactory> factory;
    FdoByteArray* array;
    void *ret = NULL;
	
	*size = 0;

    if ( geometry == NULL )
		return (char*)ret;

    factory = FdoFgfGeometryFactory::GetInstance ();
    array = factory->GetWkb ((FdoIGeometry *)geometry);
    try
    {
        *size = array->GetCount () + 4;
        ret = malloc (*size);
        memset (ret, 0, *size);
        memmove (((char*)ret) + 4, array->GetData (), *size - 4);
        array->Release ();
    }
    catch (...)
    {
        array->Release ();
    }

    return (char*)(ret);
}

void* GeometryFromExtendedWkb(void *ewkb, unsigned long size)
{
    unsigned char* l_ewkb = (unsigned char*) ewkb;
    unsigned long currentByte = 0;

    //
    // Generate FGF stream from EWKB
    //

    FdoPtr<FdoByteArray> fgfBytes = FdoByteArray::Create(size);
   FgfFromExtendedWkb(fgfBytes.p, l_ewkb, size, currentByte);

    //
    // Create FDO geometry from FGF binary data
    //
    FdoPtr<FdoFgfGeometryFactory> factory(FdoFgfGeometryFactory::GetInstance());

    FdoIGeometry* geometry = factory->CreateGeometryFromFgf(fgfBytes);

    return (void*)(geometry);
}

void FgfFromExtendedWkb(FdoByteArray*& fgfBytes,
                                       unsigned char* ewkb,
                                       unsigned long size,
                                       unsigned long& currentByte)
{
   //
    // Endian flag
    //
    unsigned char endianByte = ewkb[currentByte];
    if (eWkbNDR != endianByte)
    {
        throw FdoException::Create(FdoException::NLSGetMessage(FDO_NLSID(FDO_10_UNSUPPORTEDGEOMETRYTYPE)));
    }
    currentByte++;

    //
    // Geometry type
    //
    unsigned long ewkbFlags = *(unsigned long*)(&ewkb[currentByte]);
    unsigned long geomType = ewkbFlags & 0xff;

    // convert PostGIS curve types to FDO types
    unsigned long tarGeomType = geomType;
    if (geomType == PostGISGeometryType_CurvePolygon)
        tarGeomType = FdoGeometryType_CurvePolygon;
    else if (geomType == PostGISGeometryType_Curve || geomType == PostGISGeometryType_CompoundCurve)
        tarGeomType = FdoGeometryType_CurveString;
    else if (geomType == PostGISGeometryType_MultiCurve)
        tarGeomType = FdoGeometryType_MultiCurveString;
    else if (geomType == PostGISGeometryType_MultiSurface)
        tarGeomType = FdoGeometryType_MultiCurvePolygon;

    currentByte += sizeof(ewkbFlags);

    //
    // EWKB flags
    //

    // Dimensionality
    unsigned long geomDim = FdoDimensionality_XY;
    if (ewkbFlags & eWkbFlagZ)
        geomDim |= FdoDimensionality_Z;
    if (ewkbFlags & eWkbFlagM)
        geomDim |= FdoDimensionality_M;

    // SRID - PostGIS specific flag
    bool isSrid = (0 != (ewkbFlags & eWkbFlagSrid));

    // BBOX - PostGIS specific flag
    bool isBbox = (0 != (ewkbFlags & eWkbFlagBBox));

    if (isBbox)
    {
        throw FdoException::Create(FdoException::NLSGetMessage(FDO_NLSID(FDO_10_UNSUPPORTEDGEOMETRYTYPE)));
    }

    //
    // Generate FGF stream from EWKB
    //

    // Write Geometry type
    fgfBytes = FdoByteArray::Append(fgfBytes, sizeof(geomType), (FdoByte*)&tarGeomType);

    // Skip over Srid if present. FGF does not store it.
    if (isSrid)
    {
        currentByte += sizeof(unsigned long);
    }

    // Processing:
    // - for simple geometries types, just copy the array
    // - for multi-geometries, process sub-geometries recursively
    //
    switch(geomType)
    {
    case FdoGeometryType_MultiPoint:
    case FdoGeometryType_MultiGeometry:
    case FdoGeometryType_MultiLineString:
    case FdoGeometryType_MultiPolygon:
    case PostGISGeometryType_MultiCurve:
    case PostGISGeometryType_MultiSurface:
        {
            unsigned long numSubGeoms = *(unsigned long*)(&ewkb[currentByte]);
            fgfBytes = FdoByteArray::Append(fgfBytes, sizeof(numSubGeoms), (FdoByte*)&numSubGeoms);
            currentByte += sizeof(numSubGeoms);

            for (unsigned long i = 0;  i < numSubGeoms && currentByte < size;  i++)
            {
                // Recurse on sub-geometry.
                FgfFromExtendedWkb(fgfBytes, ewkb, size, currentByte);
            }
        }
        break;
    default:
        {
            fgfBytes = FdoByteArray::Append(fgfBytes, sizeof(geomDim), (FdoByte*)&geomDim);

            CopyGeomFromExtendedWkb(fgfBytes, ewkb, size, currentByte, geomType, geomDim);
        }
        break;
    }
}

FdoByteArray *ExtendedWkbFromGeometry (void *geometry, unsigned long srid)
{
    FdoPtr<FdoFgfGeometryFactory> factory;
    FdoPtr<FdoByteArray> fgfBytes;
    unsigned long currentByte = 0;
    unsigned long totalDims = 0;
    FdoByteArray* ewkbBytes = NULL;
	
    if ( geometry == NULL )
		return ewkbBytes;

    factory = FdoFgfGeometryFactory::GetInstance ();
    fgfBytes = factory->GetFgf((FdoIGeometry *)geometry);

    // Make ewkb buffer a bit longer to account for endian flag and srid.
    // Might not be enough for multi geometry types but the Byte array will
    // grow if needed.
    ewkbBytes = FdoByteArray::Create(fgfBytes->GetCount() + 20);
    FdoByte* fgfData = fgfBytes->GetData();
    ExtendedWkbFromFgf(ewkbBytes, fgfData, fgfBytes->GetCount(), srid, currentByte, totalDims);
    
    return ewkbBytes;
}

void ExtendedWkbFromFgf(FdoByteArray*& ewkbBytes,
                                       FdoByte* fgf,
                                       unsigned long size,
                                       unsigned long srid,
                                       unsigned long& currentByte,
                                       unsigned long& totalDims)
{
    unsigned long geomType = *(unsigned long*)(&fgf[currentByte]);
    currentByte += sizeof(geomType);

    unsigned long geomFlags = geomType;
    if (geomType == FdoGeometryType_MultiCurveString)
        geomFlags = PostGISGeometryType_MultiCurve;
    if (geomType == FdoGeometryType_MultiCurvePolygon)
        geomFlags = PostGISGeometryType_MultiSurface;

    if ( srid > 0 ) 
        geomFlags = (geomFlags | eWkbFlagSrid);

    unsigned char endianByte = eWkbNDR;
    ewkbBytes = FdoByteArray::Append(ewkbBytes, sizeof(unsigned char), (FdoByte*)&endianByte);

    // Processing:
    // - for simple geometries types, just copy the array
    // - for multi-geometries, process sub-geometries recursively
    //
    switch(geomType)
    {
    case FdoGeometryType_MultiPoint:
    case FdoGeometryType_MultiGeometry:
    case FdoGeometryType_MultiLineString:
    case FdoGeometryType_MultiPolygon:
    case FdoGeometryType_MultiCurveString:
    case FdoGeometryType_MultiCurvePolygon:
        {
            unsigned long numSubGeoms = *(unsigned long*)(&fgf[currentByte]);
            currentByte += sizeof(numSubGeoms);

            FdoInt32 flagsPosn = ewkbBytes->GetCount();
            
            ewkbBytes = FdoByteArray::Append(ewkbBytes, sizeof(geomFlags), (FdoByte*)&geomFlags);

            if ( srid > 0 ) 
                ewkbBytes = FdoByteArray::Append(ewkbBytes, sizeof(srid), (FdoByte*)&srid);

            ewkbBytes = FdoByteArray::Append(ewkbBytes, sizeof(numSubGeoms), (FdoByte*)&numSubGeoms);

            unsigned long subTotalDims = 0;

            for (unsigned long i = 0;  i < numSubGeoms && currentByte < size;  i++)
            {
                // Recurse on sub-geometry.
                ExtendedWkbFromFgf(ewkbBytes, fgf, size, srid, currentByte, subTotalDims);
            }

            FdoByte* ewkbData = ewkbBytes->GetData();

            geomFlags = (geomFlags | subTotalDims);

            *(unsigned long*)(&ewkbData[flagsPosn]) = geomFlags;
        }
        break;
    default:
        {
            unsigned long geomDim = *(unsigned long*)(&fgf[currentByte]);
            currentByte += sizeof(geomDim);

            // Dimensionality
            if ( geomDim & FdoDimensionality_Z )
            {
                geomFlags = (geomFlags | eWkbFlagZ);
                totalDims = (totalDims | eWkbFlagZ);
            }

            if ( geomDim & FdoDimensionality_M )
            {
                geomFlags = (geomFlags | eWkbFlagM);
                totalDims = (totalDims | eWkbFlagM);
            }

            FdoInt32 flagsPos = ewkbBytes->GetCount();

            ewkbBytes = FdoByteArray::Append(ewkbBytes, sizeof(geomFlags), (FdoByte*)&geomFlags);

            if ( srid > 0 ) 
                ewkbBytes = FdoByteArray::Append(ewkbBytes, sizeof(srid), (FdoByte*)&srid);

            unsigned long pgGeomType = geomType;
            CopyGeomFromFgf(ewkbBytes, fgf, size, currentByte, geomType, geomDim, pgGeomType);

            // FDO curve can be converted to PostGIS circular string or PostGIS compound curve
            // We have to use the ugly way to set the correct geometry type
            if (pgGeomType != geomType)
            {
                FdoByte* ewkbData = ewkbBytes->GetData();
                *(unsigned char*)(&ewkbData[flagsPos]) = (unsigned char)pgGeomType;
            }
        }
        break;
    }
}

void CopyGeomFromExtendedWkb(FdoByteArray*& dest,
                                       FdoByte* src,
                                       unsigned long size,
                                       unsigned long& currentByte,
                                       unsigned long geomType,
                                       unsigned long geomDim)
{
    // Valid combinations: X,Y or X,Y,Z or X,Y,M or X,Y,Z,M
    unsigned long ordinateSize =
        sizeof(double) * GetOrdinatesFromDimension(geomDim);

    switch (geomType)
    {
    case FdoGeometryType_Point:
        {
            dest = FdoByteArray::Append(dest, ordinateSize, &src[currentByte]);
            currentByte += ordinateSize;
        }
        break;
    case FdoGeometryType_LineString:
        {
            unsigned long numCoords = *(unsigned long*)(&src[currentByte]);
            dest = FdoByteArray::Append(dest, sizeof(numCoords), (FdoByte*)&numCoords);
            currentByte += sizeof(numCoords);

            unsigned long ordinatesSize =
                numCoords * ordinateSize;
            dest = FdoByteArray::Append(dest, ordinatesSize, &src[currentByte]);
            currentByte += ordinatesSize;

            //TODO: Add post-condition
        }
        break;
    case FdoGeometryType_Polygon:
        {
            unsigned long numRings = *(unsigned long*)(&src[currentByte]);
            dest = FdoByteArray::Append(dest, sizeof(numRings), (FdoByte*)&numRings);
            currentByte += sizeof(numRings);

            for (unsigned long i = 0;  i < numRings && currentByte < size;  i++)
            {
                unsigned long numCoords = *(unsigned long*)(&src[currentByte]);
                dest = FdoByteArray::Append(dest, sizeof(numCoords), (FdoByte*)&numCoords);
                currentByte += sizeof(numCoords);

                unsigned long ordinatesSize =
                    numCoords * ordinateSize;
                dest = FdoByteArray::Append(dest, ordinatesSize, &src[currentByte]);
                currentByte += ordinatesSize;
            }

        }
        break;

    case PostGISGeometryType_CurvePolygon:
    {
        unsigned long numRings = *(unsigned long*)(&src[currentByte]);
        dest = FdoByteArray::Append(dest, sizeof(numRings), (FdoByte*)&numRings);
        currentByte += sizeof(numRings);

        for (unsigned long i = 0; i < numRings && currentByte < size; i++)
        {
            // skip endian type
            currentByte++;

            unsigned long subGeomType = *(unsigned long*)(&src[currentByte]);
            currentByte += sizeof(subGeomType);

            switch (subGeomType)
            {
            case PostGISGeometryType_CompoundCurve:
            {
                PostGISCompoundCurveToFdoCurve(dest, src, size, currentByte, geomDim);
            }
            break;

            case FdoGeometryType_LineString:
            {
                unsigned long numCoords = *(unsigned long*)(&src[currentByte]);
                currentByte += sizeof(numCoords);

                // start position
                dest = FdoByteArray::Append(dest, ordinateSize, &src[currentByte]);
                currentByte += ordinateSize;

                // one segment
                unsigned long numSegs = 1;
                dest = FdoByteArray::Append(dest, sizeof(numSegs), (FdoByte*)&numSegs);

                unsigned long segGeomType = FdoGeometryComponentType_LineStringSegment;
                dest = FdoByteArray::Append(dest, sizeof(segGeomType), (FdoByte*)&segGeomType);

                // start position has been handled
                numCoords--;
                dest = FdoByteArray::Append(dest, sizeof(numCoords), (FdoByte*)&numCoords);

                unsigned long ordinatesSize = numCoords * ordinateSize;
                dest = FdoByteArray::Append(dest, ordinatesSize, &src[currentByte]);
                currentByte += ordinatesSize;
            }
            break;

            case PostGISGeometryType_Curve:
            {
                PostGISCircularStringToFdoCurve(dest, src, size, currentByte, geomDim);
            }
            break;

            default:
                throw FdoException::Create(FdoException::NLSGetMessage(FDO_NLSID(FDO_10_UNSUPPORTEDGEOMETRYTYPE)));
                break;
            }
        }
    }
    break;

    case PostGISGeometryType_Curve:
    {
        PostGISCircularStringToFdoCurve(dest, src, size, currentByte, geomDim);
    }
    break;

    case PostGISGeometryType_CompoundCurve:
    {
        PostGISCompoundCurveToFdoCurve(dest, src, size, currentByte, geomDim);
    }
    break;

    default:
        throw FdoException::Create(FdoException::NLSGetMessage(FDO_NLSID(FDO_10_UNSUPPORTEDGEOMETRYTYPE)));
    }
}

void CopyGeomFromFgf(FdoByteArray*& dest,
    FdoByte* src,
    unsigned long size,
    unsigned long& currentByte,
    unsigned long geomType,
    unsigned long geomDim,
    unsigned long& pgGeomType)
{
    // Valid combinations: X,Y or X,Y,Z or X,Y,M or X,Y,Z,M
    unsigned long ordinateSize = sizeof(double) * GetOrdinatesFromDimension(geomDim);

    switch (geomType)
    {
    case FdoGeometryType_Point:
    {
        dest = FdoByteArray::Append(dest, ordinateSize, &src[currentByte]);
        currentByte += ordinateSize;
    }
    break;

    case FdoGeometryType_LineString:
    {
        unsigned long numCoords = *(unsigned long*)(&src[currentByte]);
        dest = FdoByteArray::Append(dest, sizeof(numCoords), (FdoByte*)&numCoords);
        currentByte += sizeof(numCoords);

        unsigned long ordinatesSize =
            numCoords * ordinateSize;
        dest = FdoByteArray::Append(dest, ordinatesSize, &src[currentByte]);
        currentByte += ordinatesSize;
    }
    break;

    case FdoGeometryType_Polygon:
    {
        unsigned long numRings = *(unsigned long*)(&src[currentByte]);
        dest = FdoByteArray::Append(dest, sizeof(numRings), (FdoByte*)&numRings);
        currentByte += sizeof(numRings);

        for (unsigned long i = 0; i < numRings && currentByte < size; i++)
        {
            unsigned long numCoords = *(unsigned long*)(&src[currentByte]);
            dest = FdoByteArray::Append(dest, sizeof(numCoords), (FdoByte*)&numCoords);
            currentByte += sizeof(numCoords);

            unsigned long ordinatesSize = numCoords * ordinateSize;
            dest = FdoByteArray::Append(dest, ordinatesSize, &src[currentByte]);
            currentByte += ordinatesSize;
        }
    }
    break;

    case FdoGeometryType_CurvePolygon:
    {
        pgGeomType = PostGISGeometryType_CurvePolygon;

        unsigned long numRings = *(unsigned long*)(&src[currentByte]);
        dest = FdoByteArray::Append(dest, sizeof(numRings), (FdoByte*)&numRings);
        currentByte += sizeof(numRings);

        for (unsigned long i = 0; i < numRings && currentByte < size; i++)
        {
            unsigned char endianByte = eWkbNDR;
            dest = FdoByteArray::Append(dest, sizeof(unsigned char), (FdoByte*)&endianByte);

            unsigned long tarRingGeomType = PostGISGeometryType_CompoundCurve;
            dest = FdoByteArray::Append(dest, sizeof(tarRingGeomType), (FdoByte*)&tarRingGeomType);

            FdoCurveToPostGISCompoundCurve(dest, src, size, currentByte, geomDim);
        }
    }
    break;

    case FdoGeometryType_CurveString:
    {
        pgGeomType = PostGISGeometryType_Curve;

        // get PostGIS geometry type
        unsigned long offset = ordinateSize;
        unsigned long numSegs = *(unsigned long*)(&src[currentByte + offset]);
        offset += sizeof(numSegs);
        for (unsigned long i = 0; i < numSegs && currentByte + offset < size; i++)
        {
            unsigned long segGeomType = *(unsigned long*)(&src[currentByte + offset]);
            offset += sizeof(segGeomType);
            if (segGeomType == FdoGeometryComponentType_LineStringSegment)
            {
                pgGeomType = PostGISGeometryType_CompoundCurve;
                break;
            }
            offset += ordinateSize * 2;
        }

        if (pgGeomType == PostGISGeometryType_Curve)
        {
            unsigned long numCoords = numSegs * 2 + 1;
            dest = FdoByteArray::Append(dest, sizeof(numCoords), (FdoByte*)&numCoords);

            // start position
            dest = FdoByteArray::Append(dest, ordinateSize, &src[currentByte]);
            currentByte += ordinateSize;

            // segment count
            currentByte += sizeof(numSegs);

            for (unsigned long i = 0; i < numSegs && currentByte < size; i++)
            {
                // skip segment geometry type
                unsigned long segGeomType = *(unsigned long*)(&src[currentByte]);
                currentByte += sizeof(segGeomType);

                dest = FdoByteArray::Append(dest, ordinateSize * 2, &src[currentByte]);
                currentByte += ordinateSize * 2;
            }
        }
        else if (pgGeomType == PostGISGeometryType_CompoundCurve)
        {
            FdoCurveToPostGISCompoundCurve(dest, src, size, currentByte, geomDim);
        }
        else
            throw FdoException::Create(FdoException::NLSGetMessage(FDO_NLSID(FDO_10_UNSUPPORTEDGEOMETRYTYPE)));
    }
    break;

    default:
        throw FdoException::Create(FdoException::NLSGetMessage(FDO_NLSID(FDO_10_UNSUPPORTEDGEOMETRYTYPE)));
    }
}

unsigned long GetOrdinatesFromDimension(unsigned long dim)
{
   unsigned long value = 0;

    if (dim == FdoDimensionality_XY)
        value = 2;
    else if (dim == (FdoDimensionality_XY | FdoDimensionality_Z))
        value = 3;
    else if (dim == (FdoDimensionality_XY | FdoDimensionality_M))
        value = 3;
    else if (dim == (FdoDimensionality_XY | FdoDimensionality_Z | FdoDimensionality_M))
        value = 4;

    return value;
}

// Convert PostGIS circular string to FDO curve
void PostGISCircularStringToFdoCurve(FdoByteArray*& dest,
    FdoByte* src,
    unsigned long size,
    unsigned long& currentByte,
    unsigned long geomDim)
{
    unsigned long ordinateSize = sizeof(double) * GetOrdinatesFromDimension(geomDim);

    // numCoords must be an odd number
    unsigned long numCoords = *(unsigned long*)(&src[currentByte]);
    currentByte += sizeof(numCoords);

    // start position
    dest = FdoByteArray::Append(dest, ordinateSize, &src[currentByte]);
    currentByte += ordinateSize;

    // segment count
    unsigned long numSegs = (numCoords - 1) / 2;
    dest = FdoByteArray::Append(dest, sizeof(numSegs), (FdoByte*)&numSegs);

    for (unsigned long j = 0; j < numSegs && currentByte < size; j++)
    {
        unsigned long segGeomType = FdoGeometryComponentType_CircularArcSegment;
        dest = FdoByteArray::Append(dest, sizeof(segGeomType), (FdoByte*)&segGeomType);

        // mid point and end point
        dest = FdoByteArray::Append(dest, ordinateSize * 2, &src[currentByte]);
        currentByte += ordinateSize * 2;
    }
}

// Convert PostGIS compound curve to FDO curve
void PostGISCompoundCurveToFdoCurve(FdoByteArray*& dest,
    FdoByte* src,
    unsigned long size,
    unsigned long& currentByte,
    unsigned long geomDim)
{
    unsigned long ordinateSize = sizeof(double) * GetOrdinatesFromDimension(geomDim);

    unsigned long offset = 0;
    unsigned long numSegs = 0;
    unsigned long numCompounds = *(unsigned long*)(&src[currentByte]);
    offset += sizeof(numCompounds);

    // get segment count
    for (unsigned long i = 0; i < numCompounds && currentByte + offset < size; i++)
    {
        // skip endian type
        offset++;

        unsigned long segGeomType = *(unsigned long*)(&src[currentByte + offset]);
        offset += sizeof(segGeomType);

        unsigned long numSegCoords = *(unsigned long*)(&src[currentByte + offset]);

        switch (segGeomType)
        {
        case PostGISGeometryType_Curve:
            numSegs += (numSegCoords - 1) / 2;
            break;

        case FdoGeometryType_LineString:
            numSegs++;
            break;

        default:
            throw FdoException::Create(FdoException::NLSGetMessage(FDO_NLSID(FDO_10_UNSUPPORTEDGEOMETRYTYPE)));
            break;
        }

        offset += sizeof(numSegCoords);
        offset += numSegCoords * ordinateSize;
    }

    // start position is the first point of the first segment
    // sizeof(segment count) + sizeof(endian type) + sizeof(seg geometry type) + sizeof(seg point count) = 4 + 1 + 4 + 4 = 13
    dest = FdoByteArray::Append(dest, ordinateSize, &src[currentByte + 13]);

    // segment count
    dest = FdoByteArray::Append(dest, sizeof(numSegs), (FdoByte*)&numSegs);
    currentByte += sizeof(numSegs);

    for (unsigned long j = 0; j < numCompounds && currentByte < size; j++)
    {
        // skip endian type
        currentByte++;

        unsigned long segGeomType = *(unsigned long*)(&src[currentByte]);

        unsigned long segTarGeomType;
        if (segGeomType == PostGISGeometryType_Curve)
            segTarGeomType = FdoGeometryComponentType_CircularArcSegment;
        else if (segGeomType == FdoGeometryType_LineString)
            segTarGeomType = FdoGeometryComponentType_LineStringSegment;
        else
            throw FdoException::Create(FdoException::NLSGetMessage(FDO_NLSID(FDO_10_UNSUPPORTEDGEOMETRYTYPE)));

        dest = FdoByteArray::Append(dest, sizeof(segTarGeomType), (FdoByte*)&segTarGeomType);
        currentByte += sizeof(segTarGeomType);

        unsigned long numSubCoords = *(unsigned long*)(&src[currentByte]);
        currentByte += sizeof(numSubCoords);

        // start point has been handled
        if (j == 0)
        {
            numSubCoords--;
            currentByte += ordinateSize;
        }

        if (segTarGeomType == FdoGeometryComponentType_LineStringSegment)
        {
            dest = FdoByteArray::Append(dest, sizeof(numSubCoords), (FdoByte*)&numSubCoords);

            unsigned long ordinatesSize = numSubCoords * ordinateSize;
            dest = FdoByteArray::Append(dest, ordinatesSize, &src[currentByte]);
            currentByte += ordinatesSize;
        }
        else
        {
            unsigned long numSubCurveSegs = numSubCoords / 2;
            for (unsigned long k = 0; k < numSubCurveSegs && currentByte < size; k++)
            {
                if (k == 0)
                {
                    // skip the start point if it is the first segment, but it is not the first compound (the start
                    // point of first compound has been handled above.
                    if (j != 0)
                        currentByte += ordinateSize;
                }
                else
                {
                    dest = FdoByteArray::Append(dest, sizeof(segTarGeomType), (FdoByte*)&segTarGeomType);
                }

                unsigned long ordinatesSize = 2 * ordinateSize;
                dest = FdoByteArray::Append(dest, ordinatesSize, &src[currentByte]);
                currentByte += ordinatesSize;
            }
        }
    }
}

// Convert FDO curve to PostGIS compound curve
void FdoCurveToPostGISCompoundCurve(FdoByteArray*& dest,
    FdoByte* src,
    unsigned long size,
    unsigned long& currentByte,
    unsigned long geomDim)
{
    unsigned long ordinateSize = sizeof(double) * GetOrdinatesFromDimension(geomDim);
    unsigned char endianByte = eWkbNDR;

    // remember start point position
    unsigned long previousPos = currentByte;

    // skip start position
    currentByte += ordinateSize;

    // compound count
    unsigned long numSegs = *(unsigned long*)(&src[currentByte]);
    dest = FdoByteArray::Append(dest, sizeof(numSegs), (FdoByte*)&numSegs);
    currentByte += sizeof(numSegs);

    // one segment to one compound
    for (unsigned long i = 0; i < numSegs && currentByte < size; i++)
    {
        // segment geometry type
        unsigned long segGeomType = *(unsigned long*)(&src[currentByte]);
        currentByte += sizeof(segGeomType);

        if (segGeomType == FdoGeometryComponentType_CircularArcSegment)
        {
            // endian type
            dest = FdoByteArray::Append(dest, sizeof(unsigned char), (FdoByte*)&endianByte);

            // compound geometry type
            unsigned long compoundGeomType = PostGISGeometryType_Curve;
            dest = FdoByteArray::Append(dest, sizeof(compoundGeomType), (FdoByte*)&compoundGeomType);

            // FDO curve segment always has 3 points, start point, mid point and end point.
            unsigned long numCoords = 3;
            dest = FdoByteArray::Append(dest, sizeof(numCoords), (FdoByte*)&numCoords);

            // start point is the end point of last segment, or is the start point of the FDO curve
            dest = FdoByteArray::Append(dest, ordinateSize, &src[previousPos]);

            numCoords--;
            dest = FdoByteArray::Append(dest, ordinateSize * numCoords, &src[currentByte]);
            currentByte += ordinateSize * numCoords;

            // remember the end point position of this segment
            previousPos = currentByte - ordinateSize;
        }
        else if (segGeomType == FdoGeometryComponentType_LineStringSegment)
        {
            // endian type
            dest = FdoByteArray::Append(dest, sizeof(unsigned char), (FdoByte*)&endianByte);

            // compound geometry type
            unsigned long compoundGeomType = FdoGeometryType_LineString;
            dest = FdoByteArray::Append(dest, sizeof(compoundGeomType), (FdoByte*)&compoundGeomType);

            // point count
            unsigned long numCoords = *(unsigned long*)(&src[currentByte]);
            currentByte += sizeof(numCoords);

            // add start point if it is the first segment
            if (i == 0)
                numCoords++;

            dest = FdoByteArray::Append(dest, sizeof(numCoords), (FdoByte*)&numCoords);

            if (i == 0)
            {
                dest = FdoByteArray::Append(dest, ordinateSize, &src[previousPos]);
                numCoords--;
            }

            dest = FdoByteArray::Append(dest, ordinateSize * numCoords, &src[currentByte]);
            currentByte += ordinateSize * numCoords;

            // remember the end point position of this segment
            previousPos = currentByte - ordinateSize;
        }
        else
            throw FdoException::Create(FdoException::NLSGetMessage(FDO_NLSID(FDO_10_UNSUPPORTEDGEOMETRYTYPE)));
    }
}
